
from adverbs import * 
from array_properties import *
from array_constructors import *
from builtins import * 
from math import *
from numpy_misc import *  
from reductions import * 
from numpy_types import * 
from patchmap import *

import linalg 
from linalg import dot 
import random 