__author__  = 'Alex Rubinsteyn'
__email__   = 'alex -dot- rubinsteyn -at- gmail -dot- com'
__desc__    = 'Runtime compiler for numerical Python'
__license__     = 'BSD3'
__version__     = '0.23.2'
__website__     = 'http://www.parakeetpython.com' 

