

######################################
#        OPTIMIZER OPTIONS           #
######################################

# run LLVM optimization passes
llvm_optimize = True

# number of times to run optimizations
llvm_num_passes = 4

# run verifier over generated LLVM code?
llvm_verify = True


######################################
#         PRINTING OPTIONS           #
######################################

# print generated assembly of compiled functions
print_x86 = False

# show LLVM bytecode before optimization passes
print_unoptimized_llvm = False

# show LLVM bytecode after optimizations
# print_optimized_llvm = False 

